# -*- coding: utf-8 -*-

__author__ = "George Kadianakis, Donncha O'Cearbhaill"
__contact__ = "asn@torproject.org"
__url__ = "https://github.com/asn-d6/onionbalance"
__license__ = "GPL"

from onionbalance._version import get_versions
__version__ = get_versions()['version']
del get_versions
